{ lib ? import <nixpkgs/lib>
, common_amd64
}:

{
  overlays = common_amd64.overlays ++ [(final: prev: {

    platform_name = "kgpe";

    kernel =
      final.lib.makeOverridable (prev.kernel.override {
        config = ./linux.config;
        buildTargets = [ "bzImage" ];
      }).overrideAttrs (a: {
        postInstall = (a.postInstall or "") + ''
          cp arch/x86_64/boot/bzImage $out
        '';
      });

    userspace = prev.userspace.overrideAttrs (a: {
      postInstall = (a.postInstall or "") + ''
        mkdir -p $out/etc
        cp ${final.coreboot.src}/src/mainboard/asus/kgpe-d16/cmos.layout $out/etc/
        cp ${final.coreboot.src}/src/mainboard/asus/kgpe-d16/cmos.default $out/etc/
      '';
    });

    # microcode updates are needed only for Opteron 63xx
    #
    # TODO: check processor type at boot time and refuse to boot
    # insecure configurations? (63xx without microcode)
    microcode-blob = null;

    console-device = "ttyS0";   # the DB9 connector on the rear panel

    # This contains the default values written into the
    # battery-backed nvram when you boot with the recovery jumper
    # installed (which is something you should do the first time you
    # boot with ownerboot/coreboot).  If you want to customize these
    # values, copy `src/mainboard/asus/kgpe-d16/cmos.default` out of
    # coreboot and edit to suit your tastes.
    #
    # TODO: use a NixOS-style structuredConfig for this.
    cmos-default = null;

    coreboot = final.lib.makeOverridable (prev.coreboot.override {
      iasl = final.iasl_20180531;
      payload = "${final.kernel}/bzImage";
      fmap = ./custom.fmap;
      config = ./coreboot.config;
      coreboot-toolchain = with final.coreboot-toolchain; [ x64 i386 ];
      uart-for-console =
        if      final.console-device == "ttyS0" then 0 # the DB9 connector on the rear panel
        else if final.console-device == "ttyS1" then 1 # the IDC header on the motherboard
        else if final.console-device == null then null
        else throw "kgpe currently supports only `null` and `ttyS{0,1}` for `console-device`";
    }).overrideAttrs (a: {
      postConfigure = if final.microcode-blob != null then ''
        echo CONFIG_CPU_MICROCODE_MULTIPLE_FILES=y >> .config
      '' else ''
        echo CONFIG_CPU_MICROCODE_CBFS_NONE=y >> .config
      '';
      preBuild = final.lib.optionalString (final.microcode-blob != null) ''
        mkdir -p 3rdparty/blobs/cpu/amd/family_15h/
        ln -sfT ${final.microcode-blob} \
          3rdparty/blobs/cpu/amd/family_15h/microcode_amd_fam15h.bin
        mkdir -p 3rdparty/blobs/cpu/amd/family_10h-family_14h/
        touch 3rdparty/blobs/cpu/amd/family_10h-family_14h/microcode_amd.bin
      '' + final.lib.optionalString (final.cmos-default != null) ''
        cp ${final.cmos-default} src/mainboard/asus/kgpe-d16/cmos.default
      '';
      postInstall = (a.postInstall or "") + ''
        cp src/mainboard/asus/kgpe-d16/cmos.layout $out/
        cp src/mainboard/asus/kgpe-d16/cmos.default $out/
    '';
    });
  })];
}
